----------------------------------------------------------------------------------------
-- get resource from db
----------------------------------------------------------------------------------------
create or replace function i3dh.get_resource(p_res_id bpchar(40))
    returns table
            (
                r_id                  bigint,
                r_res_id              bpchar(40),
                r_type                text,
                r_state               integer,
                r_date_create         timestamptz,
                r_date_last_read      timestamptz,
                r_descriptor          json,
                r_endpoint_target_id  text,
                r_endpoint_names      text[],
                r_endpoint_ports      integer[],
                r_endpoint_base_paths text[],
                r_endpoint_protocols  text[],
                r_revision            integer
            )
    language sql
    stable
as
$function$
select r.id,
       r.res_id,
       r.type,
       r.state,
       r.date_create,
       r.date_last_read,
       r.descriptor,
       r.endpoint_target_id,
       re.names,
       re.ports,
       re.base_paths,
       re.protocols,
       r.revision
from i3dh.resource r
         join lateral
    (
    select array_agg(re.name)      as names,
           array_agg(re.port)      as ports,
           array_agg(re.base_path) as base_paths,
           array_agg(re.protocol)  as protocols
    from i3dh.resource_endpoints re
    where re.res_id = r.id
    ) re on true
where r.res_id = p_res_id
    ;
$function$;

alter function i3dh.get_resource(p_res_id bpchar(40)) owner to i3dhub;

-------------------------------------------------------------------------------
-- creates or reads resources from db
-------------------------------------------------------------------------------
create or replace function i3dh.get_or_create_resources(p_res_id bpchar(40)[],
                                                        p_type text,
                                                        p_date_create timestamptz)
    returns table
            (
                r_new                 bool,
                r_id                  bigint,
                r_res_id              bpchar(40),
                r_type                text,
                r_state               int4,
                r_date_create         timestamptz,
                r_date_last_read      timestamptz,
                r_descriptor          json,
                r_endpoint_target_id  text,
                r_endpoint_names      text[],
                r_endpoint_ports      integer[],
                r_endpoint_base_paths text[],
                r_endpoint_protocols  text[],
                r_revision            int4
            )
as
$body$
declare
    resource_id bpchar(40);
begin
    foreach resource_id in array p_res_id
        loop
            if not exists(select * from i3dh.resource r where r.res_id = resource_id) then
                return query
                    with ins_resource as (
                        insert into i3dh.resource (res_id, type, state, date_create, date_last_read, descriptor,
                                                   endpoint_target_id)
                            select resource_id, p_type, 1000, p_date_create, p_date_create, '{}'::json, ''::text
                            on conflict (res_id) do nothing
                            returning *)
                    select true,
                           r.id,
                           r.res_id,
                           r.type,
                           r.state,
                           r.date_create,
                           r.date_last_read,
                           r.descriptor,
                           r.endpoint_target_id,
                           null::text[],
                           null::integer[],
                           null::text[],
                           null::text[],
                           r.revision
                    from ins_resource r
                    union all
                    select false, r.*
                    from i3dh.get_resource(resource_id) r;
            else
                return query select false, r.* from i3dh.get_resource(resource_id) r;
            end if;
        end loop;
end;
$body$
    language 'plpgsql';

alter function i3dh.get_or_create_resources(bpchar(40)[], text, timestamptz) owner to i3dhub;

----------------------------------------------------------------------------------------
-- reads resources from db
----------------------------------------------------------------------------------------
create or replace function i3dh.get_resources(p_res_id bpchar(40)[])
    returns table
            (
                r_new                 bool,
                r_id                  bigint,
                r_res_id              bpchar(40),
                r_type                text,
                r_state               int4,
                r_date_create         timestamptz,
                r_date_last_read      timestamptz,
                r_descriptor          json,
                r_endpoint_target_id  text,
                r_endpoint_names      text[],
                r_endpoint_ports      integer[],
                r_endpoint_base_paths text[],
                r_endpoint_protocols  text[],
                r_revision            int4
            )
as
$body$
declare
    resource_id bpchar(40);
begin
    foreach resource_id in array p_res_id
        loop
            return query (select false,
                                 r.id,
                                 r.res_id,
                                 r.type,
                                 r.state,
                                 r.date_create,
                                 r.date_last_read,
                                 r.descriptor,
                                 r.endpoint_target_id,
                                 array_agg(re.name),
                                 array_agg(re.port),
                                 array_agg(re.base_path),
                                 array_agg(re.protocol),
                                 r.revision
                          from i3dh.resource r
                                   left join i3dh.resource_endpoints re on r.id = re.res_id
                          where r.res_id = resource_id
                          group by r.id, re.res_id);
        end loop;
    return;
end
$body$
    language 'plpgsql';

alter function i3dh.get_resources(bpchar(40)[]) owner to i3dhub;

----------------------------------------------------------------------------------------
-- reads all resources from db
----------------------------------------------------------------------------------------
create or replace function i3dh.get_all_resources()
    returns table
            (
                r_new                 bool,
                r_id                  bigint,
                r_res_id              bpchar(40),
                r_type                text,
                r_state               int4,
                r_date_create         timestamptz,
                r_date_last_read      timestamptz,
                r_descriptor          json,
                r_endpoint_target_id  text,
                r_endpoint_names      text[],
                r_endpoint_ports      integer[],
                r_endpoint_base_paths text[],
                r_endpoint_protocols  text[],
                r_revision            int4
            )
as
$body$
begin
    return query (select false,
                         r.id,
                         r.res_id,
                         r.type,
                         r.state,
                         r.date_create,
                         r.date_last_read,
                         r.descriptor,
                         r.endpoint_target_id,
                         array_agg(re.name),
                         array_agg(re.port),
                         array_agg(re.base_path),
                         array_agg(re.protocol),
                         r.revision
                  from i3dh.resource r
                           left join i3dh.resource_endpoints re on r.id = re.res_id
                  group by r.id, re.res_id);
end
$body$
    language 'plpgsql';

alter function i3dh.get_all_resources() owner to i3dhub;

----------------------------------------------------------------------------------------
-- reads expired resources from db
----------------------------------------------------------------------------------------
create or replace function i3dh.get_expired_resources(p_expire_duration interval)
    returns table
            (
                r_new                 bool,
                r_id                  bigint,
                r_res_id              bpchar(40),
                r_type                text,
                r_state               int4,
                r_date_create         timestamptz,
                r_date_last_read      timestamptz,
                r_descriptor          json,
                r_endpoint_target_id  text,
                r_endpoint_names      text[],
                r_endpoint_ports      integer[],
                r_endpoint_base_paths text[],
                r_endpoint_protocols  text[],
                r_revision            int4
            )
as
$body$
begin
    return query (select false,
                         r.id,
                         r.res_id,
                         r.type,
                         r.state,
                         r.date_create,
                         r.date_last_read,
                         r.descriptor,
                         r.endpoint_target_id,
                         array_agg(re.name),
                         array_agg(re.port),
                         array_agg(re.base_path),
                         array_agg(re.protocol),
                         r.revision
                  from i3dh.resource r
                           left join i3dh.resource_endpoints re on r.id = re.res_id
                  where r.date_last_read + p_expire_duration < now()
                  group by r.id, re.res_id);
end
$body$
    language 'plpgsql';

alter function i3dh.get_expired_resources(interval) owner to i3dhub;

----------------------------------------------------------------------------------------
-- modifies expired resource states from db
----------------------------------------------------------------------------------------
create or replace function i3dh.mark_expired_resources_for_deletion(p_expire_duration interval, p_delete_state int4, p_old_states int4[])
    returns void
as
$body$
begin
    update i3dh.resource
    set state    = p_delete_state,
        revision = revision + 1
    where state = any (p_old_states)
      and date_last_read + p_expire_duration < now();
end
$body$
    language 'plpgsql';

alter function i3dh.mark_expired_resources_for_deletion(interval, int4, int4[]) owner to i3dhub;

----------------------------------------------------------------------------------------
-- modifies all expired resources from db
----------------------------------------------------------------------------------------

create or replace function i3dh.modify_expired_resources_state(p_expire_duration interval,
                                                               p_new_state int4,
                                                               p_old_states int4[])
    returns void
as
$body$
begin
    update i3dh.resource
    set state    = p_new_state,
        revision = revision + 1
    where state = any (p_old_states)
      and date_last_read + p_expire_duration < now();
end
$body$
    language 'plpgsql';

alter function i3dh.modify_expired_resources_state(interval,int4,int4[]) owner to i3dhub;

----------------------------------------------------------------------------------------
-- reads resources from db with given state
----------------------------------------------------------------------------------------
create or replace function i3dh.get_filtered_resources_short(p_type text, p_state int4, p_limit int4, p_offset int4)
    returns table
            (
                r_res_id      bpchar(40),
                r_type        text,
                r_state       integer,
                r_date_create timestamptz
            )
as
$body$
begin
    if length(p_type) > 0 then
        return query (select r.res_id,
                             r.type,
                             r.state,
                             r.date_create
                      from i3dh.resource r
                      where r.state = p_state
                        and r.type = p_type
                      group by r.id
                      order by r.date_create desc, r.id desc
                      limit p_limit offset p_offset);
    else
        return query (select r.res_id,
                             r.type,
                             r.state,
                             r.date_create
                      from i3dh.resource r
                      where r.state = p_state
                      group by r.id
                      order by r.date_create desc, r.id desc
                      limit p_limit offset p_offset);
    end if;
end
$body$
    language 'plpgsql';

alter function i3dh.get_filtered_resources_short(text,int4,int4,int4) owner to i3dhub;

----------------------------------------------------------------------------------------
-- reads resources from db with given state
----------------------------------------------------------------------------------------
create or replace function i3dh.get_num_resources(p_type text, p_state int4)
    returns table
            (
                r_type          text,
                r_state         integer,
                r_num_resources bigint
            )
as
$body$
begin
    if length(p_type) > 0 then
        if p_state <= 0 then
            return query (select r.type,
                                 r.state,
                                 count(*)
                          from i3dh.resource r
                          where r.type = p_type
                          group by r.type, r.state);
        else
            return query (select r.type,
                                 r.state,
                                 count(*)
                          from i3dh.resource r
                          where r.state = p_state
                            and r.type = p_type
                          group by r.type, r.state);
        end if;
    else
        if p_state > 0 then
            return query (select r.type,
                                 r.state,
                                 count(*)
                          from i3dh.resource r
                          where r.state = p_state
                          group by r.type, r.state);
        else
            return query (select r.type,
                                 r.state,
                                 count(*)
                          from i3dh.resource r
                          group by r.type, r.state);
        end if;
    end if;
end
$body$
    language 'plpgsql';

alter function i3dh.get_num_resources(text,int4) owner to i3dhub;

----------------------------------------------------------------------------------------
-- reads resource type from db
----------------------------------------------------------------------------------------
create or replace function i3dh.get_resource_types()
    returns table
            (
                r_type text
            )
as
$body$
begin
    return query (select r.type
                  from i3dh.resource r
                  group by r.type);
end
$body$
    language 'plpgsql';

alter function i3dh.get_resource_types() owner to i3dhub;

----------------------------------------------------------------------------------------
-- reads resource ids from db with the given type
----------------------------------------------------------------------------------------
create or replace function i3dh.get_processed_resources_of_type(p_type text, p_limit int4, p_offset int4)
    returns table
            (
                r_res_id bpchar(40),
                r_state  integer
            )
as
$body$
begin
    return query (select r.res_id, r.state
                  from i3dh.resource r
                  where r.state >= 3000
                    and r.type = p_type
                  order by r.date_create desc, r.id desc
                  limit p_limit offset p_offset);
end
$body$
    language 'plpgsql';

alter function i3dh.get_processed_resources_of_type(text,int4,int4) owner to i3dhub;

----------------------------------------------------------------------------------------
-- modifies all resources with old states to the new state
----------------------------------------------------------------------------------------
create or replace function i3dh.modify_resource_states(p_new_state int4, p_old_states integer[])
    returns void
as
$body$
begin
    update i3dh.resource
    set state    = p_new_state,
        revision = revision + 1
    where state = any (p_old_states);
end
$body$
    language 'plpgsql';

alter function i3dh.modify_resource_states(int4,integer[]) owner to i3dhub;

----------------------------------------------------------------------------------------
-- updates values of a resource
----------------------------------------------------------------------------------------
create or replace function i3dh.update_resource(p_res_id bpchar(40),
                                                p_state int4,
                                                p_type text,
                                                p_date_last_read timestamptz,
                                                p_descriptor json,
                                                p_endpoint_target_id text,
                                                p_endpoint_names text[],
                                                p_endpoint_ports integer[],
                                                p_endpoint_base_paths text[],
                                                p_endpoint_protocols text[],
                                                p_parameters json)
    returns boolean
as
$body$
declare
    updated_id bigint;
begin
    updated_id := (select id from i3dh.resource where res_id = p_res_id);

    -- update resource values
    update i3dh.resource
    set state              = coalesce(p_state, state),
        type               = coalesce(p_type, type),
        date_last_read     = coalesce(p_date_last_read, date_last_read),
        descriptor         = coalesce(p_descriptor, descriptor),
        endpoint_target_id = coalesce(p_endpoint_target_id, endpoint_target_id),
        revision           = revision + 1
    where id = updated_id;

    if not found then
        return false;
    end if;

    -- update or insert parameters
    if not p_parameters is null then
        insert into i3dh.resource_parameters(res_id, parameters)
        values (updated_id, p_parameters)
        on conflict (res_id) do update set parameters = p_parameters;
    end if;

    -- see if we need to update endpoints
    if cardinality(p_endpoint_names) <> 0 or cardinality(p_endpoint_ports) <> 0 or
       cardinality(p_endpoint_base_paths) <> 0 or cardinality(p_endpoint_protocols) <> 0 then
        -- delete old endpoints
        delete
        from i3dh.resource_endpoints
        where res_id = updated_id;

        -- write new endpoints
        for i in array_lower(p_endpoint_names, 1) .. array_upper(p_endpoint_names, 1)
            loop
                insert into i3dh.resource_endpoints(res_id, name, port, base_path, protocol)
                values (updated_id,
                        coalesce(p_endpoint_names[i], ''),
                        coalesce(p_endpoint_ports[i], 0),
                        coalesce(p_endpoint_base_paths[i], ''),
                        coalesce(p_endpoint_protocols[i], ''));
            end loop;
    end if;

    return true;
end
$body$
    language 'plpgsql';

alter function i3dh.update_resource(bpchar(40), int4, text, timestamptz, json, text, text[], integer[], text[], text[], json) owner to i3dhub;

----------------------------------------------------------------------------------------
-- batch updates last read date for given resources
----------------------------------------------------------------------------------------
create or replace function i3dh.update_resource_last_read(p_res_id bpchar(40)[])
    returns void
    language sql
    volatile
as
$body$
update i3dh.resource
set date_last_read = now()
where res_id = any (p_res_id);
$body$;

alter function i3dh.update_resource_last_read(p_res_id bpchar(40)[]) owner to i3dhub;

----------------------------------------------------------------------------------------
-- batch updates resource states
----------------------------------------------------------------------------------------
create or replace function i3dh.update_resource_state_batch(p_res_id bpchar(40)[],
                                                            p_new_state int4[],
                                                            p_old_revisions int4[])
    returns void
as
$body$
begin
    for i in array_lower(p_res_id, 1) .. array_upper(p_res_id, 1)
        loop
            update i3dh.resource
            set state    = p_new_state[i],
                revision = revision + 1
            where res_id = p_res_id[i]
              and revision = p_old_revisions[i];
        end loop;
    return;
end
$body$
    language 'plpgsql';

alter function i3dh.update_resource_state_batch(bpchar(40)[], int4[], int4[]) owner to i3dhub;

----------------------------------------------------------------------------------------
-- atomic update only if not in any of the forbidden states, returning success
----------------------------------------------------------------------------------------
create or replace function i3dh.set_resource_state_atomic(p_res_id bpchar(40),
                                                          p_new_state int4,
                                                          p_forbidden_old_states int4[])
    returns boolean
as
$body$
begin
    update i3dh.resource
    set state    = p_new_state,
        revision = revision + 1
    where res_id = p_res_id
      and not state = any (p_forbidden_old_states);
    return found;
end
$body$
    language 'plpgsql';

alter function i3dh.set_resource_state_atomic(bpchar(40), int4, int4[]) owner to i3dhub;

----------------------------------------------------------------------------------------
-- deletes resources
----------------------------------------------------------------------------------------
create or replace function i3dh.delete_resources(p_res_id bpchar(40)[])
    returns void
as
$body$
declare
    resource_id bpchar(40);
    deleted_id  bigint;
begin
    foreach resource_id in array p_res_id
        loop
            deleted_id := (select id from i3dh.resource where res_id = resource_id);
            delete from i3dh.resource_endpoints re where re.res_id = deleted_id;
            delete from i3dh.resource_parameters rp where rp.res_id = deleted_id;
            delete from i3dh.resource r where r.id = deleted_id;
        end loop;
    return;
end
$body$
    language 'plpgsql';

alter function i3dh.delete_resources(bpchar(40)[]) owner to i3dhub;

----------------------------------------------------------------------------------------
-- get resource children short
----------------------------------------------------------------------------------------

create or replace function i3dh.get_resource_children_short(p_res_id bpchar(40))
    returns table
            (
                r_res_id      bpchar(40),
                r_type        text,
                r_state       integer,
                r_date_create timestamptz
            )
as
$body$
declare
begin
    return query with recursive children as (select rc.parent_id,
                                                    rc.child_id
                                             from i3dh.resource r,
                                                  i3dh.resource_children rc
                                             where r.res_id = p_res_id
                                               and rc.parent_id = r.id
                                             union
                                             select rc2.parent_id,
                                                    rc2.child_id
                                             from i3dh.resource_children rc2
                                                      inner join children c on c.child_id = rc2.parent_id)
                 select r.res_id,
                        r.type,
                        r.state,
                        r.date_create
                 from children ch,
                      i3dh.resource r
                 where ch.child_id = r.id;
end
$body$
    language 'plpgsql';

alter function i3dh.get_resource_children_short(bpchar(40)) owner to i3dhub;

----------------------------------------------------------------------------------------
-- set resource children
----------------------------------------------------------------------------------------

create or replace function i3dh.set_resource_children(p_parent_id bpchar(40), p_child_ids bpchar(40)[])
    returns void
as
$body$
declare
    v_parent_id bigint;
    v_child_id  bpchar(40);
begin
    v_parent_id := (select r.id from i3dh.resource as r where r.res_id = p_parent_id);
    delete from i3dh.resource_children as rc where rc.parent_id = v_parent_id;

    foreach v_child_id in array p_child_ids
        loop
            insert into i3dh.resource_children
            values (v_parent_id, (select r.id from i3dh.resource as r where r.res_id = v_child_id));
        end loop;
end
$body$
    language 'plpgsql';

alter function i3dh.set_resource_children(bpchar(40), bpchar(40)[]) owner to i3dhub;

create or replace function i3dh.query_etag_resource_params(p_resource_id bpchar(40))
    returns table
            (
                r_type           text,
                r_etag           text,
                r_input_uri      text,
                r_service_params text
            )
    language sql
as
$body$
select r.type,
       (r.descriptor #>> '{source,eTag}'),
       (rp.parameters #>> '{Params,inputUri}'),
       (rp.parameters #>> '{Params,original}')
from i3dh.resource r,
     i3dh.resource_parameters rp
where rp.res_id = r.id
  and r.res_id = p_resource_id;
$body$;

alter function i3dh.query_etag_resource_params(p_resource_id bpchar(40)) owner to i3dhub;
